import shutil
import os
import argparse
import unittest
import io

from tokenizer.tokenizer import Tokenizer

class TestTokenizer(unittest.TestCase):

    MODEL_DIR = '/tmp/stanza_models'
    
    def setUp(self):
        self.args = {
            'lang': 'it',
            'dir': self.MODEL_DIR,
            'verbose': True
        }

    def tearDown(self):
        shutil.rmtree(self.MODEL_DIR)
        
    def test_download_resources(self):
        self.assertTrue(not os.path.exists(self.MODEL_DIR))
        tokenizer = Tokenizer(**self.args)
        self.assertTrue(os.path.exists(self.args['dir']) and not os.path.isfile(self.args['dir']))
        self.assertTrue(os.path.exists(os.path.join(self.args['dir'], 'tokenizer', self.args['lang'])))
        self.assertTrue(os.path.exists(os.path.join(self.args['dir'], 'tokenizer', self.args['lang'], 'tokenize')))
    
    def test_tokenize(self):
        tokenizer = Tokenizer(**self.args)
        sentences = tokenizer.predict('Ha chiamato il dr. Rossi.Vuole salutarti.')
        self.assertEqual(len(sentences), 2)

    def test_corpus_load(self):
        tokenizer = Tokenizer(**self.args)
        sin = io.StringIO("Un corazziere contro Scalfaro. L'attore le disse baciami o torno a riprendermelo.")
        
        for line in tokenizer.format(tokenizer.predict(sin.read())):
            if line and not line.startswith('#'):
                assert len(line.split('\t')) == 10, line



# -*- coding: utf-8 -*-

import torch.nn as nn
from ..modules.dropout import SharedDropout


class MLP(nn.Module):
    r"""
    Applies a linear transformation together with :class:`~torch.nn.LeakyReLU` activation to the incoming tensor:
    :math:`y = \mathrm{LeakyReLU}(x A^T + b)`

    Args:
        n_in (~torch.Tensor):
            The size of each input feature.
        n_out (~torch.Tensor):
            The size of each output feature.
        dropout (float):
            If non-zero, introduce a :class:`SharedDropout` layer on the output with this dropout ratio. Default: 0.
    """

    def __init__(self, n_in, n_out, dropout=0):
        super().__init__()

        self.n_in = n_in
        self.n_out = n_out
        self.linear = nn.Linear(n_in, n_out)
        self.activation = nn.LeakyReLU(negative_slope=0.1)
        self.dropout = SharedDropout(p=dropout)

        self.reset_parameters()

    def __repr__(self):
        s = f"n_in={self.n_in}, n_out={self.n_out}"
        if self.dropout.p > 0:
            s += f", dropout={self.dropout.p}"

        return f"{self.__class__.__name__}({s})"

    def reset_parameters(self):
        nn.init.orthogonal_(self.linear.weight)
        nn.init.zeros_(self.linear.bias)

    def forward(self, x):
        r"""
        Args:
            x (~torch.Tensor):
                The size of each input feature is `n_in`.

        Returns:
            A tensor with the size of each output feature `n_out`.
        """

        x = self.linear(x)
        x = self.activation(x)
        x = self.dropout(x)

        return x



# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
from torch.nn.utils.rnn import pack_padded_sequence


class CharLSTM(nn.Module):
    r"""
    CharLSTM aims to generate character-level embeddings for tokens.
    It summerizes the information of characters in each token to an embedding using a LSTM layer.

    Args:
        n_char (int):
            The number of characters.
        n_embed (int):
            The size of each embedding vector as input to LSTM.
        n_out (int):
            The size of each output vector.
        pad_index (int):
            The index of the padding token in the vocabulary. Default: 0.
    """

    def __init__(self, n_chars, n_word_embed, n_out, pad_index=0):
        super().__init__()

        self.n_chars = n_chars
        self.n_word_embed = n_word_embed
        self.n_out = n_out
        self.pad_index = pad_index

        # the embedding layer
        self.embed = nn.Embedding(num_embeddings=n_chars,
                                  embedding_dim=n_word_embed)
        # the lstm layer
        self.lstm = nn.LSTM(input_size=n_word_embed,
                            hidden_size=n_out//2,
                            batch_first=True,
                            bidirectional=True)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.n_chars}, {self.n_embed}, n_out={self.n_out}, pad_index={self.pad_index})"

    def forward(self, x):
        r"""
        Args:
            x (~torch.Tensor): ``[batch_size, seq_len, fix_len]``.
                Characters of all tokens.
                Each token holds no more than `fix_len` characters, and the excess is cut off directly.
        Returns:
            ~torch.Tensor:
                The embeddings of shape ``[batch_size, seq_len, n_out]`` derived from the characters.
        """

        # [batch_size, seq_len, fix_len]
        mask = x.ne(self.pad_index)
        # [batch_size, seq_len]
        lens = mask.sum(-1)
        char_mask = lens.gt(0)

        # [n, fix_len, n_word_embed]
        x = self.embed(x[char_mask])
        x = pack_padded_sequence(x, lens[char_mask], True, False)
        x, (h, _) = self.lstm(x)
        # [n, fix_len, n_out]
        h = torch.cat(torch.unbind(h), dim=-1)
        # [batch_size, seq_len, n_out]
        embed = h.new_zeros(*lens.shape, self.n_out)
        embed = embed.masked_scatter_(char_mask.unsqueeze(-1), h)

        return embed



# -*- coding: utf-8 -*-

import diaparser
from diaparser.parsers import Parser

model = 'en_ewt-electra-base'

def test_parse():
    sentence = ['The', 'dog', 'chases', 'the', 'cat', '.']
    parser = Parser.load(model)
    parser.predict([sentence], prob=True)


def test_bert():
    sentence = '''
        No, it wasn't Black Monday.
        But while the New York Stock Exchange didn't fall apart Friday as the Dow Jones Industrial Average
        plunged 190.58 points - most of it in the final hour - it barely managed to stay this side of chaos.
        Some "circuit breakers" installed after the October 1987 crash failed their first test, traders say,
        unable to cool the selling panic in both stocks and futures.
        The 49 stock specialist firms on the Big Board floor - the buyers and sellers of last resort
        who were criticized after the 1987 crash - once again couldn't handle the selling pressure.
        Big investment banks refused to step up to the plate to support the beleaguered floor traders
        by buying big blocks of stock, traders say.
        Heavy selling of Standard & Poor's 500-stock index futures in Chicago relentlessly beat stocks downward.
        Seven Big Board stocks - UAL, AMR, BankAmerica, Walt Disney, Capital Cities/ABC,
        Philip Morris and Pacific Telesis Group - stopped trading and never resumed.
        The finger-pointing has already begun. "The equity market was illiquid.
        Once again {the specialists} were not able to handle the imbalances on the floor of the New York Stock Exchange,"
        said Christopher Pedersen, senior vice president at Twenty-First Securities Corp.
        Countered James Maguire, chairman of specialists Henderson Brothers Inc.:
        "It is easy to say the specialist isn't doing his job.
        When the dollar is in a free-fall, even central banks can't stop it.
        Speculators are calling for a degree of liquidity that is not there in the market."
        Many money managers and some traders had already left their offices early Friday afternoon on a warm autumn day -
        because the stock market was so quiet.
        Then in a lightning plunge,
        the Dow Jones industrials in barely an hour surrendered about a third of their gains this year,
        chalking up a 190.58-point, or 6.9%, loss on the day in gargantuan trading volume.
        Final-hour trading accelerated to 108.1 million shares, a record for the Big Board.
        At the end of the day, 251.2 million shares were traded.
        The Dow Jones industrials closed at 2569.26.
        The Dow's decline was second in point terms only to the 508-point Black Monday crash that occurred Oct. 19, 1987.
        '''
    parser = Parser.load(model)
    parser.predict(sentence, lang='en', prob=True)



# -*- coding: utf-8 -*-

from .affine import Biaffine
from .bert import BertEmbedding
from .char_lstm import CharLSTM
from .dropout import IndependentDropout, SharedDropout
from .lstm import LSTM
from .mlp import MLP
from .scalar_mix import ScalarMix

__all__ = ['MLP', 'BertEmbedding', 'Biaffine', 'CharLSTM',
           'IndependentDropout', 'LSTM', 'ScalarMix', 'SharedDropout']



"""
Query the DiaParser model catalog.
"""

import os
import requests
from tqdm import tqdm
from pathlib import Path
import json
import hashlib
import logging
import subprocess

from .. import __version__, __models_version__

logger = logging.getLogger('diaparser')

RELEASE = 'v1.0'
DOWNLOAD_URL = f'https://github.com/Unipisa/diaparser/releases/download/{RELEASE}'

DEFAULT_CATALOG_URL = DOWNLOAD_URL
DEFAULT_CATALOG_VERSION = __models_version__

# set home dir for default
HOME_DIR = str(Path.home())
CACHE_DIR = os.path.join(HOME_DIR, '.cache/diaparser')


def get_md5(path):
    """
    Get the MD5 value of a path.
    """
    data = open(path, 'rb').read()
    return hashlib.md5(data).hexdigest()


def file_exists(path, md5):
    """
    Check if the file at `path` exists and match the provided md5 value.
    """
    return os.path.exists(path) and get_md5(path) == md5


def download_file(url, path):
    """
    Download a URL into a file as specified by `path`.
    """
    verbose = logger.level in [0, 10, 20]
    r = requests.get(url, stream=True)
    with open(path, 'wb') as f:
        file_size = int(r.headers.get('content-length'))
        default_chunk_size = 131072
        desc = 'Downloading ' + url
        with tqdm(total=file_size, unit='B', unit_scale=True,
                  disable=not verbose, desc=desc) as pbar:
            for chunk in r.iter_content(chunk_size=default_chunk_size):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    pbar.update(len(chunk))


def request_file(url, path, md5=None):
    """
    A complete wrapper over download_file() that also make sure the directory of
    `path` exists, and that a file matching the md5 value does not exist.
    Args:
        url: of file to retrieve.
        path: where to store result.
        md5: the ecpected md5 of the file.
    """

    Path(path).parent.mkdir(parents=True, exist_ok=True)
    if file_exists(path, md5):
        logger.info(f'File exists: {path}.')
        return
    download_file(url, path)
    assert(not md5 or file_exists(path, md5))


def url_ok(url):
    """
    Checks that a given URL is reachable.
    :param url: A URL
    :rtype: bool
    """
    return requests.head(url).ok


def available_models(lang, dir, catalog_url, catalog_version):
    """Download list of available models for :param lang:."""

    logger.debug('Downloading catalog file...')
    # make request
    catalog_path = os.path.join(dir, 'catalog.json')
    request_file(
        f'{catalog_url}/catalog-{catalog_version}.json',
        catalog_path)
    # unpack results
    try:
        catalog = json.load(open(catalog_path))
    except:
        raise Exception(
            'Cannot load model catalog. Please check your network connection, '
            'or provided resource URL and resource version.'
        )
    if lang not in catalog:
        raise Exception(f'Unsupported language: {lang}.')
    models = catalog[lang]
    if 'alias' in models:
        alias = models['alias']
        logger.info(f'"{alias}" is an alias for "{lang}"')
        corpora = models[alias]
    return models


def select(name=None, lang='en', corpus=None, bert=None,
           dir=CACHE_DIR,
           verbose=None,
           catalog_url=DEFAULT_CATALOG_URL,
           catalog_version=DEFAULT_CATALOG_VERSION,
           **kwargs):
    """
    Determines which model to download.
    If `name` is provided, the model with that name is returned.
    If just `lang` is specifiled, it selects the default model for the languagae.
    If `lang` and `corpus` are specified, it returns the model for the given language/corpus pair.
    Args:
        name (str): model name.
        lang (str): the language of the model.
        corpus (str): the corpus of the model.
    Returns:
        the URL from where to download the model.
    """

    # set global logging level
    logging_level = 'INFO' if verbose else 'ERROR'
    logger.setLevel(logging_level)

    # Check for a model with the given name.
    url = f'{catalog_url}/{name}'
    if url_ok(url):
        return url

    models = available_models(lang, dir, catalog_url, catalog_version)
    
    if corpus not in models:
        logger.info(f'Using {models["default"]} as corpus for {lang}')
        corpus = models.get("default", None)
    model = models[corpus].get('parse', None) if corpus in models else None
    if model:
        return f'{catalog_url}/{model}'
    return None


def available_processors(lang='en', corpus=None,
                         dir=CACHE_DIR,
                         verbose=None,
                         catalog_url=DEFAULT_CATALOG_URL,
                         catalog_version=DEFAULT_CATALOG_VERSION,
                         **kwargs):
    """
    Determines which processors (parse, tokenize, mwt) are available.
    If just `lang` is specifiled, it selects the default corpus for the languagae.
    If `lang` and `corpus` are specified, it returns the model for the given language/corpus pair.
    Args:
        lang (str): the language of the model.
        corpus (str): the corpus for the chosen language.
    Returns:
        dict of {'tokenize': tok_url, 'mwt': mwt_url} from where to download the preprocessors models.
    """
    models = available_models(lang, dir, catalog_url, catalog_version)
    if corpus not in models:
        logger.info(f'Using {models["default"]} as corpus for {lang}')
        corpus = models.get('default', None)
    return models.get(corpus, {})


def download_processors(lang, processors, dir,
                        download_url=DOWNLOAD_URL):
    paths = {}
    for proc,model in processors.items():
        try:
            path = os.path.join(dir, lang, proc, f'{model}.pt')
            request_file(f'{download_url}/{model}.pt', path)
            paths[proc + '_model_path'] = path
        except KeyError as e:
                raise Exception(
                    f'Cannot find the following processor and model name combination: '
                    f'{proc}, {model}. Please check if you have provided the correct model name.'
                ) from e
    return paths


def upload(path, owner='Unipisa', repo='diaparser', version='v1.0', token=''):
    name = os.path.basename(path)
    GH_ASSET = f"https://uploads.github.com/repos/{owner}/{repo}/releases/{version}/assets?name={name}"
    curl = f'curl --data-binary @"{path}" -H "Authorization: token {token}" -H "Content-Type: application/zip" {GH_ASSET}'
    subprocess.run(curl.split())



# -*- coding: utf-8 -*-

import os
from datetime import datetime, timedelta

import torch
import torch.distributed as dist
from .. import parsers
from tokenizer.tokenizer import Tokenizer
from ..catalog import select
from ..utils import Config, Dataset
from ..utils.field import Field, BertField
from ..utils.logging import init_logger, logger
from ..utils.metric import Metric
from ..utils.parallel import DistributedDataParallel as DDP
from ..utils.parallel import is_master
from torch.optim import Adam
from torch.optim.lr_scheduler import ExponentialLR


class Parser():

    MODEL = None

    def __init__(self, args, model, transform):
        self.args = args
        self.model = model
        self.transform = transform

    def train(self, train, dev, test,
              buckets=32,
              batch_size=5000,
              lr=2e-3,
              mu=.9,
              nu=.9,
              epsilon=1e-12,
              clip=5.0,
              decay=.75,
              decay_steps=5000,
              epochs=5000,
              patience=100,
              verbose=True,
              **kwargs):
        r"""
        Args:
            lr (float): learnin rate of adam optimizer. Default: 2e-3.
            mu (float): beta1 of adam optimizer. Default: .9.
            nu (float): beta2 of adam optimizer. Default: .9.
            epsilon (float): epsilon of adam optimizer. Default: 1e-12.
            buckets (int): number of buckets. Default: 32.
            epochs (int): number of epochs to train: Default: 5000.
            patience (int): early stop after these many epochs. Default: 100.
        """

        args = self.args.update(locals())
        init_logger(logger, verbose=args.verbose)

        self.transform.train()
        if dist.is_initialized():
            args.batch_size = args.batch_size // dist.get_world_size()
        logger.info(f"Load the datasets\n"
                    f"{'train:':6} {train}\n"
                    f"{'dev:':6} {dev}\n")
        train = Dataset(self.transform, args.train, **args)
        train.build(args.batch_size, args.buckets, True, dist.is_initialized())
        logger.info(f"{'train:':6} {len(train):5} sentences, "
                    f"{len(train.loader):3} batches, "
                    f"{len(train.buckets)} buckets")
        dev = Dataset(self.transform, args.dev)
        dev.build(args.batch_size, args.buckets)
        logger.info(f"{'dev:':6} {len(dev):5} sentences, "
                    f"{len(dev.loader):3} batches, "
                    f"{len(train.buckets)} buckets")
        if args.test:
            test = Dataset(self.transform, args.test)
            test.build(args.batch_size, args.buckets)
            logger.info(f"{'test:':6} {len(test):5} sentences, "
                        f"{len(test.loader):3} batches, "
                        f"{len(train.buckets)} buckets")
        else:
            test = None

        logger.info(f"Model\n{self.model}\n")
        if dist.is_initialized():
            self.model = DDP(self.model,
                             device_ids=[dist.get_rank()],
                             find_unused_parameters=True)
        self.optimizer = Adam(self.model.parameters(),
                              args.lr,
                              (args.mu, args.nu),
                              args.epsilon)
        self.scheduler = ExponentialLR(self.optimizer, args.decay**(1/args.decay_steps))

        elapsed = timedelta()
        best_e, best_metric = 1, Metric()

        for epoch in range(1, args.epochs + 1):
            start = datetime.now()

            logger.info(f"Epoch {epoch} / {args.epochs}:")
            self._train(train.loader)
            loss, dev_metric = self._evaluate(dev.loader)
            logger.info(f"{'dev:':6} - loss: {loss:.4f} - {dev_metric}")
            if test:
                loss, test_metric = self._evaluate(test.loader)
                logger.info(f"{'test:':6} - loss: {loss:.4f} - {test_metric}")

            t = datetime.now() - start
            # save the model if it is the best so far
            if dev_metric > best_metric:
                best_e, best_metric = epoch, dev_metric
                if is_master():
                    self.save(args.path)
                logger.info(f"{t}s elapsed (saved)\n")
            else:
                logger.info(f"{t}s elapsed\n")
            elapsed += t
            if epoch - best_e >= args.patience:
                break

        logger.info(f"Epoch {best_e} saved")
        logger.info(f"{'dev:':6} - {best_metric}")
        if test:
            loss, metric = self.load(args.path)._evaluate(test.loader)
            logger.info(f"{'test:':6} - {metric}")
        logger.info(f"{elapsed}s elapsed, {elapsed / epoch}s/epoch")

    def evaluate(self, data, buckets=8, batch_size=5000, **kwargs):
        args = self.args.update(locals())
        init_logger(logger, verbose=args.verbose)

        self.transform.train()
        logger.info("Loading the data")
        dataset = Dataset(self.transform, data)
        dataset.build(args.batch_size, args.buckets)
        logger.info(f"\n{dataset}")

        logger.info("Evaluating the dataset")
        start = datetime.now()
        loss, metric = self._evaluate(dataset.loader)
        elapsed = datetime.now() - start
        logger.info(f"loss: {loss:.4f} - {metric}")
        logger.info(f"{elapsed}s elapsed, {len(dataset)/elapsed.total_seconds():.2f} Sents/s")

        return loss, metric

    def predict(self, data, pred=None, buckets=8, batch_size=5000, prob=False, **kwargs):
        args = self.args.update(locals())
        init_logger(logger, verbose=args.verbose)

        self.transform.eval()
        if args.prob:
            self.transform.append(Field('probs'))
        if hasattr(args, 'text') and args.text \
           or hasattr(args, 'lang') and args.lang: # PATCH: back compatibility
            self.transform.reader = Tokenizer(args.text, dir=args.cache_dir).reader()

        logger.info("Loading the data")
        dataset = Dataset(self.transform, data)
        dataset.build(args.batch_size, args.buckets)
        logger.info(f"\n{dataset}")

        logger.info("Making predictions on the dataset")
        start = datetime.now()
        preds = self._predict(dataset.loader)
        elapsed = datetime.now() - start

        for name, value in preds.items():
            setattr(dataset, name, value)
        if pred is not None and is_master():
            logger.info(f"Saving predicted results to {pred}")
            self.transform.save(pred, dataset.sentences)
        logger.info(f"{elapsed}s elapsed, {len(dataset) / elapsed.total_seconds():.2f} Sents/s")

        return dataset

    def _train(self, loader):
        raise NotImplementedError

    @torch.no_grad()
    def _evaluate(self, loader):
        raise NotImplementedError

    @torch.no_grad()
    def _predict(self, loader):
        raise NotImplementedError

    @classmethod
    def build(cls, path, **kwargs):
        raise NotImplementedError

    @classmethod
    def load(cls, name_or_path, cache_dir=os.path.expanduser('~/.cache/diaparser'), **kwargs):
        r"""
        Loads a parser from a pretrained model.

        Args:
            name_or_path (str):
                - a string with the shortcut name of a pretrained parser listed in ``resource.json``
                  to load from cache or download, e.g., ``'en_ptb.electra-base'``.
                - a path to a directory containing a pre-trained parser, e.g., `./<path>/model`.
            cache_dir (str):
                Directory where to cache models. The default value is `~/.cache/diaparser`.
            kwargs (dict):
                A dict holding the unconsumed arguments that can be used to update the configurations and initiate the model.

        Examples:
            >>> parser = Parser.load('en_ewt.electra-base')
            >>> parser = Parser.load('./ptb.biaffine.dependency.char')
        """

        args = Config(**locals())
        args.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        if os.path.exists(name_or_path):
            state = torch.load(name_or_path)
        else:
            url = select(name=name_or_path, **kwargs)
            if url is None:
                raise Exception(f'Could not find a model matching name {name_or_path}')
            verbose = kwargs.get('verbose', True)
            state = torch.hub.load_state_dict_from_url(url, model_dir=cache_dir,
                                                       progress=verbose)
        cls = getattr(parsers, state['name'])
        args = state['args'].update(args)
        model = cls.MODEL(**args)
        model.load_pretrained(state['pretrained'])
        model.load_state_dict(state['state_dict'], False)
        model.to(args.device)
        transform = state['transform']
        if args.feat == 'bert':
            tokenizer = BertField.tokenizer(args.bert)
            transform.FORM[1].tokenize = tokenizer.tokenize
        return cls(args, model, transform)

    def save(self, path):
        model = self.model
        if hasattr(model, 'module'):
            model = self.model.module
        args = model.args
        args.pop('Parser')      # dont save parser class object
        state_dict = {k: v.cpu() for k, v in model.state_dict().items()}
        pretrained = state_dict.pop('pretrained.weight', None)
        if args.feat == 'bert':
            tokenize = self.transform.FORM[1].tokenize  # save it
            self.transform.FORM[1].tokenize = None
        state = {'name': type(self).__name__,
                 'args': args,
                 'state_dict': state_dict,
                 'pretrained': pretrained,
                 'transform': self.transform}
        torch.save(state, path)
        if args.feat == 'bert':
            self.transform.FORM[1].tokenize = tokenize  # restore
